#include "RobustLHSR.h"

//------------------------------------------------------------------------------
// Class RobustMEM
//-------------------------------------------------
// Public Members

virtual ~RobustLHSR()
{
  unsigned size = archive.size();
  for (unsigned i = 0; i < size; i++)
    delete archive[i];
}

void RobustLHSR::updateTotalGen(unsigned& totalGen, unsigned currGen, unsigned evaluations, unsigned lambda)
{
  totalGen = currGen + (evaluations - evalsUsed) / (lambda * sampleSetSize);  
}

//------------------------------------------------------------------------------
// Class RobustMEM
//-------------------------------------------------
// Protected Members

void RobustLHSR::robustSimulate(Individual* I, int* xnom, int numVar) // FIXME: noise not on diameters, but on demands and pipe roughness coefficients
{
#ifdef DEBUG
  cout << "RobustLHSR::robustSimulate()" << endl;
#endif
  
  vector<double>* F = getF();
  if (numF == 0)
    numF = F->size();

  /* Approximate Latin Hypercube reference set by selecting suitable ArchiveTuples */ 
  archiveSelection(I, xnom, numVar);

  /* Evaluate candidate Points for extra sampling */
  Individual tmpI;	  
  unsigned size = candidates.size();
  for (unsigned i = 0; i < size; i++)
  {
    /* Simulate using object variables in candidate Point */	  
    tmpI->D = candidates[i]->D;
    simulate(tmpI, NULL, numVar, NULL);
    evalsUsed++;
    
    /* Add new ArchiveTuple to archive and 'assign' it to Point in reference set */
    selectedArchiveTuples.push_back(archive.size());
    archive.push_back(new ArchiveTuple((*candidates[i]), F));
  }

  /* Determine effective fitness */
  vector<double> F_eff(numF, 0);
  double weight, totalWeight = 0;
  for (unsigned i = 0; i < sampleSetSize; i++)
  {
    /* Retrieve probability of I->D (or xnom) being perturbed into archive[selectedArchiveTuples[i]]->D */	  
    weight = weightD(archive[selectedArchiveTuples[i]]->D, I, xnom);
    totalWeight += weight;

    for (int j = 0; j < numF; j++)
      F_eff[j] += weight * archive[selectedArchiveTuples[i]]->F[j];  
  }

  for (int j = 0; j < numF; j++)
    F->at(j) = F_eff[j] /= totalWeight;

  if (I != NULL)
    I->F = F_eff;
  
  cleanupAfterSim();
}

/* Approximate Latin Hypercube reference set by selecting suitable ArchiveTuples */ 
void RobustLHSR::archiveSelection(Individual* I, int* xnom, int numVar)
{
  /* Generate Latin Hypercube reference set based on supplied object variables and Robust::noise() */	
  LHS(I, xnom, numVar);

  /* Determine the need for extra sampling per Point in the reference set */
  unsigned archiveIndex;
  for (unsigned refSetIndex = 0; refSetIndex < sampleSetSize; refSetIndex++)
  {
    /* Retrieve ArchiveTuple closest to Point refSetIndex in the reference set */
    archiveIndex = getClosestArchiveTuple(refSetIndex);

    /* Check that ArchiveTuple archiveIndex is not closer to another Point in the reference set */
    if (noCloserRefPoint(refSetIndex, archiveIndex))   // If check succeeds, 'assign' ArchiveTuple to Point 
      selectedArchiveTuples.push_back(archiveIndex);   // refSetIndex
    else                                               // Otherwise, add point refsetindex                                          
      candidates.push_back(referenceSet[refSetIndex]); // to candidates for extra sampling
  }
}

/* Generate Latin Hypercube reference set based on supplied object variables and stdInputNoise; 
 * for discrete nominal object variables 
 */	
void RobustLHSR::LHS(Individual* I, int* xnom, int numVar) 
{
  
}

/* Retrieve ArchiveTuple closest to Point refSetIndex in the reference set */
unsigned RobustLHSR::getClosestArchiveTuple(unsigned refSetIndex)
{
  unsigned size = archive.size();
  double distance, distanceClosest = -1;
  unsigned archiveIndexClosest;
  for (unsigned archiveIndex = 0; archiveIndex < size; archiveIndex++)
  {
    distance = EuclidianDistanceD(refSetIndex, archiveIndex);	  
    if (distance < distanceClosest || distanceClosest == -1)
    {
      archiveIndexClosest = archiveIndex;
      distanceClosest = distance;
    }  
  }

  return archiveIndexClosest;
}

/* Calculate Euclidian distance between Point refSetIndex in the reference set and 
 * ArchiveTuple archiveIndex in the archive; for discrete nominal object variables
 */
double RobustLHSR::EuclidianDistanceD(unsigned refSetIndex, unsigned archiveIndex)
{
  unsigned size = referenceSet[refSetIndex]->D->size();	
  double distance = 0;
  for (unsigned i = 0; i < size; i++)
    distance += pow((double) (referenceSet[refSetIndex]->D[i] - archive[archiveIndex]->D[i]), 2);
  
  return sqrt(distance);
}

/* Check that ArchiveTuple archiveIndex is not closer to another Point in the reference set */
bool RobustLHSR::noCloserRefPoint(refSetIndex, archiveIndex)
{
  double distance = EuclidianDistanceD(refSetIndex, archiveIndex)	
  for (unsigned i = 0; i < sampleSetSize; i++)
  {
    if (i == refSetIndex)
      continue;

    /* Closer Point found in reference set */
    if (EuclidianDistanceD(i, archiveIndex) < distance)
      return false;    
  }

  /* No closer Point found in reference set */
  return true;
}

/* Retrieve probability of I->D (or xnom) being perturbed into archive[selectedArchiveTuples[i]]->D */	 
double RobustLHSR::weightD(vector<int>& D, Individual* I, int* xnom)
{

}

void RobustLHSR::cleanupAfterSim()
{
  int size = referenceSet.size();	
  for (int i = 0; i < size; i++)
    delete referenceSet[i];
  referenceSet.clear();
    
  selectedArchiveTuples.clear();

  candidates.clear();
}


