#ifndef WDN_
#define WDN_

#include "mies/SerialES/MixedIntegerES.h"
#ifdef MEM
#include "RobustMEM.h"
#endif
#ifdef LHSR
#include "RobustLHSR.h"
#endif
#include "toolkit.h"

#define EN_ERR_INPUT     302 
#define EN_ERR_OUTPUT    303

#define HEADER_DIAMETERS "[DIAMETERS]"
#define HEADER_MISC      "[MISC]"

struct Pipe
{
  float length;
    
  int beginNode,
      endNode;

  float roughness;
};

struct Network
{	
      //! Number of pipes
  int m, 
      //! Number of internal nodes
      n, 
      //! Number of reservoir nodes 
      r;

               //! New pipes in the network (parallel to existing ico parallel expansion problem) 
  vector<Pipe> pipe,      
               //! Existing pipes in the network (only ico parallel expansion problem) 
               existPipe;

	        //! Elevation of each internal node
  vector<float> elevation,
                //! Demand at each internal node  
	        Q;      
};

struct Diameter
{
  float size,
        price;
};

struct DiameterTable:public vector<Diameter> 
{ 
  int lookupDiameterCode(float diameterSize);
};

class WDN
#ifdef MEM
 :public RobustMEM
#endif
#ifdef LHSR
 :public RobustLHSR
#endif
{
  public:
    WDN(string simInputFileName, string simOutputFileName, double cvDemand_, double cvRoughness_, unsigned sampleSetSize_, bool LH_ = true)
#ifdef MEM
    :RobustMEM(sampleSetSize_, LH_)
#endif
#ifdef LHSR
    :RobustLHSR(sampleSetSize_)
#endif
    {  
      cvDemand    = cvDemand_;
      cvRoughness = cvRoughness_;
      initialize(simInputFileName, simOutputFileName);
    }

    void simulate(int* xnom, int nnom) { simulate(NULL, xnom, nnom, NULL); }

    bool isParallelExp() { return parallelExp; }
    
    vector<double>* getF() { return &F; }

    void getNadirIdeal(vector<double>& nadir, vector<double>& ideal, int numVar);

    virtual ~WDN() { close(); }

  protected:
    vector<double> F;

    Network network;
    DiameterTable diameterTable;

    float pressureZero,
          pressureReq,
	  //! Unit conversion constant for diameter size 
          convDiamSize;

         /** \brief  Indicates parallel expansion problem, introducing zero diameter,
                     and compensating for existing pipes */
    bool parallelExp;

           //! Coefficient of variation for demands		      
    double cvDemand, 
           //! Coefficient of variation for Hazen-Williams roughness coefficients
	   cvRoughness;

    //! Initialize EPANET simulator 
    void initialize(string simInputFileName, string simOutputFileName);

    void simulate(Individual* I, int* xnom, int numVar, double* noise); 

    //! f1: Investment Cost -> min 
    double f1(Individual* I, int* xnom); 

    //! f2: Weighted Demand Supply Ratio -> MAX
    double f2(vector<float>& pressure); 

    //! f3: System Entropy -> MAX  
    double f3(vector<float>& flow);  

    //! Demands of n nodes 
    int* getNumNoiseNorm() { return &network.n; }

    //! Roughness of m pipes
    int* getNumNoiseUnif() { if (!parallelExp) return NULL; return &network.m; }

    //! Shutdown EPANET simulator
    void close();
};

#endif
