#include "APriori_MIES.h"

//------------------------------------------------------------------------------
// Class APriori_MIES
//-------------------------------------------------
// Protected Members

/* Select mu Individuals from O and optionally P to form new population P */
void APriori_MIES::select()
{
#ifdef DEBUG
  cout << "APriori_MIES::select()" << endl;
#endif

  /* Partition population into fronts */
  vector<Individual*> Q;
  int size = lambda;
  for (unsigned i = 0; i < lambda; i++)
  {
    Q.push_back(O[i]);
    O[i] = NULL;
  }

  /* Select Individuals from parent population P that do not exceed maximum life span */
  for (unsigned i = 0; i < mu; i++)
  {
    if (P[i]->age <= kappa)
    {
      Q.push_back(P[i]);
      P[i] = NULL;
      size++;
    } 
  }

  /* Generate vector with indices for Q */
  vector<int> indices;
  for (int i = 0; i < size; i++)
    indices.push_back(i);
  
  /* Cleanup unused Individuals from parent population P */
  for (unsigned i = 0; i < mu; i++)
  {
    delete P[i];
    P[i] = NULL;
  }

  /* Determine new population */
  P.clear();
  quickSort(indices, 0, size-1, &Q, selectFunctionActual); // Sort Individuals in descending order 

  /* Add best mu Individuals to population */
  int k;
  for (unsigned i = (direction[selectFunctionActual] == 1) ? 0 : size-1, j = 0; j < mu; j++)
  {
    k = indices[i + j*direction[selectFunctionActual]];
    P.push_back(Q[k]);
    Q[k] = NULL;
  }

  /* Cleanup unused Individuals */
  for (int i = 0; i < size; i++)
    delete Q[i];

  /* Determine bestF value */
  determineBestF();
}

/* Sort Individuals on objective, in descending order */
void APriori_MIES::quickSort(vector<int>& indices, int top, int bottom, vector<Individual*>* Q, int objective)
{
#ifdef DEBUG
  cout << "APriori_MIES::quickSort(), sorting on objective " + itos(objective) << endl;
#endif

  /* top = subscript of beginning of vector being considered
   * bottom = subscript of end of vector being considered
   */
  int middle;
  if (top < bottom)
  {
    middle = partition(indices, top, bottom, Q, objective);
    quickSort(indices, top, middle, Q, objective); // Sort top partition
    quickSort(indices, middle+1, bottom, Q, objective); // Sort bottom partition
  }
}

int APriori_MIES::partition(vector<int>& indices, int top, int bottom, vector<Individual*>* Q, int objective)
{
#ifdef DEBUG
  cout << "APriori_MIES::partition(), top=" << top << ", bottom=" << bottom << endl;
#endif

  int temp, 
      i = top - 1, 
      j = bottom + 1;

  double x = Q->at(indices[top])->F[objective];

  do
  {
    do     
    {
      j--;
    } while (x > Q->at(indices[j])->F[objective]);

    do  	
    {
      i++;
    } while (x < Q->at(indices[i])->F[objective]);

    if (i < j)
    {     
      /* Switch elements at positions i and j */
      temp = indices[i];    
      indices[i] = indices[j];
      indices[j] = temp;
    }
  } while (i < j);    
  
  /* Return middle index */
  return j;           
}

/* Determine bestF value for current population */
void APriori_MIES::determineBestF()
{
#ifdef DEBUG
  cout << "APriori_MIES::determineBestF()" << endl;
#endif

  vector<int> indices;
  for (unsigned i = 0; i < mu; i++)
    indices.push_back(i);

  /* Sort the indices in descending order depending on the selectFunctionActual value of the individual it points to */
  quickSort(indices, 0, mu-1, &P, selectFunctionActual); 
   
  int k;
  for (unsigned i = (direction[selectFunctionActual] == 1) ? 0 : mu-1, j = 0; j < mu; j++)
  {
    k = indices[i + j*direction[selectFunctionActual]];

    if (!(P[k]->feasible)) // Allow only fitness value from feasible solution 
      continue;

    bestF[selectFunctionActual] = P[k]->F[selectFunctionActual];
    break;

    bestF[selectFunctionActual] = -1; // No feasible solutions in population, set to error value -1
  }
}

/* In case of convergence study (i.e., with known optima supplied) determine whether optima have been reached */
bool APriori_MIES::optimumReached()
{
#ifdef DEBUG
  cout << "APriori_MIES::optimumReached()" << endl;
#endif

  if (generations == USHRT_MAX) // Convergence study
  {
    if (direction[selectFunctionActual] == 1)
    {
      if (bestF[selectFunctionActual] >= optimalF[selectFunctionActual])
	return true;
    }
    else
    {
      if (bestF[selectFunctionActual] <= optimalF[selectFunctionActual])
	return true;
    }
  }

  return false;
}

