#ifndef MIES_
#define MIES_

#include "global.h"

class Individual;

/** \brief  Abstract class */
class MixedIntegerES
{
  public:
    MixedIntegerES(unsigned n_r_, unsigned n_z_, unsigned n_d_, unsigned n_sigma_r_, unsigned n_sigma_z_, unsigned n_prob_, unsigned n_f_, 
		    unsigned mu_, unsigned rho_, unsigned kappa_, unsigned lambda_,  
		      vector<double>& lBound_r_, vector<double>& uBound_r_, vector<int>& lBound_z_, vector<int>& uBound_z_, 
		        vector<int>& lBound_d_, vector<int>& uBound_d_, double prob_min_, double prob_max_,
			  bool disRec_r_, bool disRec_z_, bool disRec_sigma_r_, bool disRec_sigma_z_, bool disRec_prob_,
			    bool selfAdaptation_, vector<int>& direction_, 
			      unsigned evaluations_, vector<double>& optimalF_,
                                double initialSigma_r_, double initialSigma_z_, double initialProb_,
			          ifstream* initialPopFile_, int randomSeed_, bool feedback_);
    virtual ~MixedIntegerES();

    /** \brief  Execute the MIES */
    virtual void run(); 

  protected: 
    unsigned n_r, n_z, n_d, n_sigma_r, n_sigma_z, n_prob, n_f; 
    unsigned mu, rho, kappa, lambda;
    vector<double> lBound_r, uBound_r; vector<int> lBound_z, uBound_z, lBound_d, uBound_d; double prob_min, prob_max;
    bool disRec_r, disRec_z, disRec_sigma_r, disRec_sigma_z, disRec_prob;
    bool selfAdaptation; vector<double> maxSigma_r, maxSigma_z; vector<int> direction; vector<double> bestF; 
    unsigned evaluations, generations, currGen; vector<double> optimalF; 
    double initialSigma_r, initialSigma_z, initialProb;
    ifstream* initialPopFile; int randomSeed; bool feedback; bool debug; time_t id;
    
    /** \brief  Parents */
    vector<Individual*> P; 

    /** \brief  Offspring */
    vector<Individual*> O;  

    /** \brief  Initialize Individuals in population P with uniform randomly 
      *         chosen values within various variable and parameter domains 
      *	 	or read previously generated population in from file */
    virtual void initialize(); 

    /** \brief  Determine fitness values for population P or O */
    virtual void evaluate(); 

    /** \brief  Run simulator using the object variables contained in Individual I */
    virtual void simulate(Individual* I) = 0; 

    /** \brief  Write current population P to file */
    virtual void writePop(time_t id) = 0; 

    /** \brief  Increase age of Individuals in population P by one */
    void increaseAge(); 

    /** \brief  Create lambda offspring by recombination and mutation */
    virtual void recombineMutate(); 

    /** \brief  Create lambda offspring by mutation of a single (recombined) parent */
    void recombineMutateSingleParent(); 

    /** \brief  Output stepsizes */
    virtual void writeStep(time_t id) {}

    void applyRecomb(Individual* I, vector<int>& parents);

    void applyMut(Individual* I);

    /** \brief  Reflect double x back into domain [a,b]; b must be strictly greater than a */
    double transformR(double x, double a, double b); 

    /** \brief  Reflect integer x back into domain [a,b]; b must be strictly greater than a */
    int transformZ(int x, int a, int b); 

    /** \brief  Fixed convergence for supplied stepsize or probability */
    double fixedConvergence(double x, double threshold, double varPart); 

    /** \brief  Select mu Individuals from O and optionally P to form new population P */
    virtual void select() = 0; 

    /** \brief  Log time that was needed to complete generation */
    virtual void writeLog(time_t elapsed, time_t id) = 0;

    /** \brief  In case of convergence study (i.e., with known optima supplied) 
                determine whether optima have been reached */
    virtual bool optimumReached() = 0; 
};

class Individual 
{
  public:
    Individual() { feasible = true; age = 0; numDominating = 0; }
    Individual(unsigned n_r, unsigned n_z, unsigned n_d, unsigned n_sigma_r, unsigned n_sigma_z, unsigned n_prob, unsigned n_f,
                vector<double>& lBound_r, vector<double>& uBound_r, vector<int>& lBound_z, vector<int>& uBound_z, 
		  vector<int>& lBound_d, vector<int>& uBound_d, double prob_min, double prob_max,
		    double initialSigma_r, double initialSigma_z, double initialProb);
    Individual(vector<double>& R_, vector<int>& Z_, vector<int>& D_, 
	        vector<double>& S_r_, vector<double>& S_z_, vector<double>& Prob_,
		  vector<double>& F_, bool feasible_);
    Individual(Individual* original);
    ~Individual() {}    
    void dump();                                                                                         

    /* Object variables */

    /** \brief  Real-valued */
    vector<double> R; 

    /** \brief  Integer */
    vector<int> Z;  

    /** \brief  Discrete */
    vector<int> D; 

    /* Strategy parameters */

    /** \brief  Stepsizes for real-valued variables */
    vector<double> S_r;

    /** \brief  Stepsizes for integer variables */
    vector<double> S_z; 

    /** \brief  Mutation probabilities for discrete variables */
    vector<double> Prob; 
    
    /** \brief  Fitness values */
    vector<double> F;

    bool feasible;

    unsigned age;

    /** \brief  Number of Individuals in population by which this Individual is dominated */
    unsigned numDominating;
};

#endif
